CON

  _CLKMODE = XTAL1 + PLL16X
  _XINFREQ = 5_000_000

  'Parallax Serial Terminal Control Character Constants
  HM =  1 'HM: Home Cursor       
  PC =  2 'PC: Position Cursor In X, Y
  ML =  3 'ML: Move Cursor Left          
  MR =  4 'MR: Move Cursor Right         
  MU =  5 'MU: Move Cursor Up          
  MD =  6 'MD: Move Cursor Down
  BP =  7 'BP: Beep Speaker
  BS =  8 'BS: Backspace
  TB =  9 'TB: Tab
  LF = 10 'LF: Line Feed
  CE = 11 'CE: Clear To End Of line     
  CB = 12 'CB: Clear Lines Below
  CR = 13 'CR: Carriage Return          
  PX = 14 'PX: Position Cursor In X         
  PY = 15 'PY: Position Cursor In Y
  CS = 16 'CS: Clear Screen
  SP = 32 'Space  

  'Transmission And Reception Constants
  STX = 2 'Start Of Text
  ETX = 3 'End Of Text
  ENQ = 5 'Enquire the RESULT of the transmission
  ACK = 6 'Positive RESULT of a successful transmission
  NAK = 21 'Negative RESULT of an unsuccessful transmission
  US = 31 'Data block seperator which is used for parsing transmission data
  COL = 58 'Colon delimiter useful for seperating strings into seperate fields

  'CRC Formatting Constants
  CRC_HEAD = %1110_0000 '4 bit tag, 4 bit payload
  CRC_MASK4 = %0000_1111 'Payload mask
  CRC_TAIL = %10_000000 '2 bit tag, 6 bit payload
  CRC_MASK6 = %00_111111 'Payload mask

  'Buffer size for reception buffers
  MESSAGE_BUFFER_SIZE = 24
  CRC_BUFFER_SIZE = 8

  'Default delimiter
  DELIMITER = COL
  
VAR
  LONG SenderStack[32]
  LONG NewTxDataString
  
  BYTE MessageBuffer[MESSAGE_BUFFER_SIZE]
  BYTE CRCBuffer[CRC_BUFFER_SIZE]
  LONG  DataField1
  LONG  DataField2

OBJ
  Terminal : "FullDuplexSerial"
  Sender : "FullDuplexSerial"
  Receiver : "FullDuplexSerial"

PUB Main

  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  'The parser expects two numbers, with each number being no
  'larger than 2,147,483,647, and the numbers being seperated by
  'a colon delimiter.  If you require something a little more
  'fancy, you will have to create different parsing routines.
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  NewTxDataString  := STRING("2147483647:2147483647")
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

  Terminal.Start(31, 30, 0, 115_200)
  Receiver.Start(0, 1, 0, 9600)
  
  WAITCNT(CLKFREQ * 2 + CNT)

  Terminal.Tx(CS)
  Terminal.Rxflush

  Terminal.Tx(CR)
  Terminal.Str(STRING("Prop2PropSim", CR))
  Terminal.Tx(CR)
  Terminal.Tx(CR)
  Terminal.Str(STRING("The STRING's were placed in the PUB blocks instead of the DAT block", CR))
  Terminal.Str(STRING("to help beginners to understand the hows, wheres, whats, and whys.", CR))
  Terminal.Tx(CR)

  COGNEW(SenderComm, @SenderStack)

  MonitorForIncomingData   

PUB MonitorForIncomingData | Index, Char, CRCExpected, CRCReceived, NumberOfRxAttempts, RxCounter

  REPEAT UNTIL Receiver.Rx == STX

  Terminal.Str(STRING("MessageBuffer filling has started!", CR))

  Index := 0

  REPEAT WHILE (Char := Receiver.Rx) <> US
    MessageBuffer[Index] := Char
    Index++

  MessageBuffer[Index~] := 0                            ' terminate string, clear index
  
  Terminal.Str(STRING("MessageBuffer filling has ended!", CR))
  Terminal.Str(STRING("CRCBuffer filling has started!", CR))

  REPEAT WHILE (Char := Receiver.Rx) <> ETX
    CRCBuffer[Index] := Char
    Index++

  Terminal.Str(STRING("CRCBuffer filling has ended!", CR))
  Terminal.Str(STRING("ETX has been found!", CR))

  Terminal.Str(STRING("################################################################", CR))
  Terminal.Str(STRING("Received data without the CRC value:  "))
  Terminal.Str(@MessageBuffer)
  Terminal.Tx(CR)
  Terminal.Str(STRING("#################################################################", CR))

  Terminal.Str(STRING("Parsing MessageBuffer into seperate fields!", CR))

  ParseReceivedData(@MessageBuffer, @DataField1, @DataField2)  

  Terminal.Str(STRING("Value For Field 1: "))
  Terminal.Dec(DataField1)
  Terminal.Tx(CR)
  Terminal.Str(STRING("Value For Field 2: "))
  Terminal.Dec(DataField2)
  Terminal.Tx(CR)

  Terminal.Str(STRING("Finished with parsing MessageBuffer into seperate fields!", CR))  
  
  BYTEMOVE(@CRCReceived, @CRCBuffer, 3)
  Terminal.Str(STRING("Received CRC value: "))
  Terminal.dec(decodeCRC(CRCReceived))
  Terminal.Tx(CR)
  
  CRCExpected := ComputeCRC(@MessageBuffer, STRSIZE(@MessageBuffer))

  Terminal.Str(STRING("Expected CRC value: "))
  Terminal.Dec(CRCExpected)
  Terminal.Tx(CR)  

  IF CRCExpected == decodeCRC(CRCReceived)
    Terminal.Str(STRING("CRC comparisons were successful!", CR))    
    Terminal.Str(STRING("ACK will be sent to Sender indicating a successful transmission!", CR))
    Receiver.Tx(ACK)

  ELSE
    Terminal.Str(STRING("CRC comparisons were unsuccessful!", CR))    
    Terminal.Str(STRING("NAK will be sent to Sender indicating an unsuccessful transmission!", CR))
    Receiver.Tx(NAK)

PUB SenderComm

  Sender.Start(1, 0, 0, 9600)
  WAITCNT(CLKFREQ * 4 + CNT)

  IF NOT \ SendDataStringWithCRC
    Terminal.Str(STRING("Multiple transmission attempts were unsuccessfull!", CR))
    
PUB SendDataStringWithCRC | CRC, Char, NumberOfTxAttempts, TxCounter, OldTxDataString

  REPEAT
    IF OldTxDataString <> NewTxDataString
      OldTxDataString := NewTxDataString

      TxCounter := 0
      NumberOfTxAttempts := 3

      REPEAT NumberOfTxAttempts

        TxCounter++        

        CRC := EncodeCRC(ComputeCRC(NewTxDataString, STRSIZE(NewTxDataString)))

        'Send Start Of Text message to the Receiver
        Sender.Tx(STX)
        { Format and send main data content between here and the next comment }

        Sender.Str(NewTxDataString)
        Sender.Tx(US)
      
        REPEAT 3
      
          Sender.Tx(CRC)
          CRC >>= 8

        { Main data content should have been formatted and sent by this point }
        'Send End Of Text message to the Receiver
        Sender.Tx(ETX)

        IF Char := Sender.RxTime(1_000)

          IF Char == ACK

            Terminal.Str(STRING("ACK received, valid data was transmitted!", CR))
            Terminal.Str(STRING("Waiting for new data!", CR))
            Terminal.Tx(CR)

            QUIT

          IF Char == NAK

            IF TxCounter == NumberOfTxAttempts

              Terminal.Str(STRING("All attempts to send valid data have failed!  Now aborting!", CR))

              ABORT FALSE

            ELSE

              Terminal.Str(STRING("NAK received, invalid data was transmitted!", CR))

          IF Char == -1

            IF TxCounter == NumberOfTxAttempts

              Terminal.Str(STRING("All attempts to retrieve acknowledgement of valid data have failed!  Now aborting!", CR))

              ABORT FALSE

            ELSE

              Terminal.Str(STRING("No acknowledgement of valid data has been received!  Still Checking!", CR))

PUB ParseReceivedData(ReceivedDataPointer, Field1Pointer, Field2Pointer)

  ReceivedDataPointer := Trim(ReceivedDataPointer)
  LONG[Field1Pointer] := StrToDec(ReceivedDataPointer)

  ReceivedDataPointer += FindChar(DELIMITER, ReceivedDataPointer, 11)
  ReceivedDataPointer := Trim(ReceivedDataPointer)      
  LONG[Field2Pointer] := StrToDec(ReceivedDataPointer)

PUB Trim(StrPointer) | Char

  REPEAT

    Char := BYTE[StrPointer]

    IF ((Char => "0") and (Char =< "9")) 

      RETURN StrPointer
      
    ELSE

      ++StrPointer                    

PUB FindChar(Char, StrPointer, StrLength) | Index

  RESULT := -1

  REPEAT Index FROM 0 TO StrLength - 1
  
    IF (BYTE[StrPointer++] == Char)
    
      RESULT := Index
      
      QUIT

PUB StrToDec(StrPointer) | Char

  RESULT := 0

  REPEAT

    Char := BYTE[StrPointer++]

    IF ((Char => "0") and (Char =< "9"))
    
      RESULT := (RESULT * 10) + (Char - "0")
      
    ELSE
    
      QUIT

PUB ComputeCRC(DataPointer, SizeOfBytes)
  
  REPEAT SizeOfBytes
  
    RESULT ^= (BYTE[DataPointer++] << 8)
    
    REPEAT 8
    
      RESULT <<= 1
      
      IF RESULT & $10000
      
        RESULT ^= $1021
        
  RESULT &= $FFFF

PRI EncodeCRC(CrcDecoded) : CrcEncoded

  CrcEncoded.BYTE[0] := CRC_TAIL | CRC_MASK6 &  CrcDecoded         ' %----------------10cccccc
  CrcEncoded.BYTE[1] := CRC_TAIL | CRC_MASK6 & (CrcDecoded >> 6)   ' %--------10cccccc--------
  CrcEncoded.BYTE[2] := CRC_HEAD | CRC_MASK4 & (CrcDecoded >> 12)  ' %1110cccc----------------
  
PRI DecodeCRC(CrcEncoded) : CrcDecoded

  CrcDecoded |= CRC_MASK6 & CrcEncoded                             ' %----------cccccc
  CrcDecoded |= (CONSTANT(CRC_MASK6 <<  8) & CrcEncoded) >> 2      ' %----cccccc------
  CrcDecoded |= (CONSTANT(CRC_MASK4 << 16) & CrcEncoded) >> 4      ' %cccc------------
  